<!DOCTYPE style-sheet PUBLIC "-//James Clark//DTD DSSSL Style
Sheet//EN" [

<!ENTITY useful PUBLIC "-//O'Reilly//NONSGML Procedures::Useful DSSSL
Procedures//EN" NDATA DSSSL>

<!ENTITY amp CDATA "&">
<!ENTITY lt CDATA "<">
]>

<style-specification use="useful">

;; Refer to this stylesheet as
;; PUBLIC
;;    "-//O'Reilly//NONSGML Procedures::MIF Generation Procedures//EN"
;; 
;; You must define certain objects:
;; 
;; blocks is a list of strings, each the GI of an element whose
;;    construction rule will create a MIF Para.
;; inlines is a list of strings, each the GI of an element whose
;;    construction rule will create a MIF Font, so that the current
;;    Font can be restored after it is processed.
;; (make-columns) is a function for creating MIF column definitions
;;    from your SGML table.
;; (make-footnotes) is a function for generating the content of a MIF
;;    Notes section.
;; (make-table-content) is a function for creating the content of the
;;    MIF table.
;; (mode tables) is a mode for generating the content of tables, in
;;    the MIF Tbls section.  A default construction rule is provided.
;; nobreaks is a list of strings, each the name of a MIF FTag for
;;    which there should be no internal line breaks.  Necessary only
;;    if the pi-handler uses it; if used, it must be constructed by
;;    appending strings to mif-nobreaks.
;; (pi-handler this-node parafmt stringfmt) is a procedure that
;;    processes a node of class 'pi and returns a flow object.
;; preserves is a list of strings, each the name of a GI for which
;;    every new-line should mean a new paragraph, as for code
;;    listings.
;; (sdata-handler this-node parafmt stringfmt) is a procedure that
;;    processes a node of class 'sdata and returns a flow object.
;; table-formats is a list of strings, each the name of a TblFormat
;;    you want to be able to use in your MIF document.  Without a
;;    catalog at the top of the MIF file, Frame will crash opening it.
;; xref-formats is a list of strings, each the name of an XRefFormat
;;    you want to be able to use in your MIF document.  I suspect that
;;    Frame will crash without this list, as it does with the
;;    TblFormats.


;; Files are created as entity flow objects.
(declare-flow-object-class entity
			   "UNREGISTERED::James Clark//Flow Object Class::entity")

;; formatting-instruction flow object class; everything is done in
;; here.
(declare-flow-object-class formatting-instruction
			   "UNREGISTERED::James Clark//Flow Object Class::formatting-instruction")

;; all-element-number is needed for generating MIF Unique tags, Marker
;; text, and XRefSrcText.
(define all-element-number
  (external-procedure "UNREGISTERED::James Clark//Procedure::all-element-number"))

;; (line-break) creates a "soft" return within a paragraph.
(define (line-break)
  (make formatting-instruction
	data: "'>
   &lt;Char HardReturn>
  >
  &lt;ParaLine 
   &lt;String `"))

;; (make-footnote) creates the MIF tagging for a footnote within the
;; Notes section.
(define (make-footnote)
  (sosofo-append  (make formatting-instruction
			data: (string-append "  &lt;FNote 
   &lt;ID "
					     (number->string (all-element-number))
					     ">
"))
		  (process-children)
		  (make formatting-instruction
			data: "  > # end of FNote
")))

;; (make-footnote-ref snl) creates the MIF for a reference to a
;; footnote at node snl
(define (make-footnote-ref snl)
  (make formatting-instruction
	data: (string-append "'>
   &lt;FNote "
			     (number->string (all-element-number snl))
			     ">
   &lt;String `")))

;; (make-mif-start):
;; starts the MIFFile
;; includes all known table formats first
;; then all known xref formats
;; then all tables
;; then all footnotes
;; then other content
(define (make-mif-start)
  (sosofo-append (make formatting-instruction
		       data: (string-append "&lt;MIFFile 5.00> # Generated by Jade 1.0
&lt;TblCatalog 
"
					    (let t-f-loop ((tblfmts table-formats))
						 (if (null? tblfmts)
						     ""
						     (string-append " &lt;TblFormat 
  &lt;TblTag `"
								    (list-ref tblfmts
									      0)
								    "'>
 > # end of TblFormat
"
								    (if (null? (list-tail tblfmts
											  1))
									""
								        (t-f-loop (list-tail tblfmts
											     1))))))
					    "> # end of TblCatalog
&lt;XRefFormats 
"
					    (let x-f-loop ((xrffmts xref-formats))
						 (if (null? xrffmts)
						     ""
						     (string-append " &lt;XRefFormat 
  &lt;XRefName `"
								    (list-ref xrffmts
									      0)
								    "'>
 > # end of XRefFormat
"
								    (if (null? (list-tail xrffmts
											  1))
									""
								        (x-f-loop (list-tail xrffmts
											     1))))))
					    "> # end of XRefFormats
&lt;Tbls 
"))
		 ;; process tables
		 (with-mode tables
			    (process-children))
		 (make formatting-instruction
		       data: "> # end of Tbls
&lt;TextFlow 
 &lt;TFTag `A'>
 &lt;TFAutoConnect Yes>
 &lt;Notes 
")
		 ;; process footnotes
		 (make-footnotes)
		 (make formatting-instruction
		       data: " > # end of Notes
")))

;; (make-mif-end) ends the MIF file
(define (make-mif-end)
  (make formatting-instruction
	data: "> # end of TextFlow
# end of MIFFile
"))

;; (make-para fmtname is-title) creates a MIF paragraph tagged as type
;; fmtname and handles its content; if is-title, it also creates a
;; Marker.
(define (make-para fmtname
		   is-title)
  (sosofo-append (make formatting-instruction
		       data: (string-append (parafmt fmtname)
					    "  &lt;ParaLine 
"))
		 (if is-title
		     (sosofo-append (make formatting-instruction
					  data: (string-append "   &lt;Marker 
    &lt;MType 9>
    &lt;MText `"
							       (number->string (all-element-number (current-node)))
							       ": "
							       fmtname
							       ": "))
				    (with-mode text-only
					       (process-text fmtname "text-only"))
				    (make formatting-instruction
					  data: (string-append "'>
    &lt;Unique "
							       (number->string (all-element-number (current-node)))
							       ">
   > # end of Marker
")))
		     (empty-sosofo))
		 (make formatting-instruction
		       data: "   &lt;String `")
		 (process-text fmtname "")
		 (make formatting-instruction
		       data: "'>
  >
 > # end of Para
")))

;; (make-ruled-para fmtname) creates a MIF paragraph tagged as type
;; fmtname, but sets PgfBotSeparator to `' if it's not the last
;; paragraph in its container.
(define (make-ruled-para fmtname)
  (if (last-sibling?)
      (make-para fmtname
		 #f)
      (sosofo-append (make formatting-instruction
			   data: (string-append (parafmt fmtname)
						"  &lt;Pgf 
   &lt;PgfBotSeparator `'>
  > # end of Pgf
  &lt;ParaLine 
   &lt;String `"))
		     (process-text fmtname "")
		     (make formatting-instruction
			   data: "'>
  >
 > # end of Para
"))))

;; (make-start-xref xreffmt tgtfmt content-gi) creates the beginning
;; of an XRef; call it from the target node.  The xreffmt is used for
;; the XRefName, the tgtfmt is the PgfFmt of the target's native
;; occurrence, and the content-gi is the gi of the child of the target
;; to use for the text (e.g., the title of the target container).
(define (make-start-xref xreffmt tgtfmt content-gi)
  (sosofo-append (make formatting-instruction
		       data: (string-append "    &lt;XRefName `"
					    xreffmt
					    "'>
    &lt;XRefSrcText `"
					    (number->string (all-element-number (get-children-by-type (list content-gi)
												      (current-node))))
					    ": "
					    tgtfmt
					    ": "))
		 (with-mode text-only
			    (process-first-descendant content-gi))
		 (make formatting-instruction
		       data: (string-append "'>
    &lt;XRefSrcIsElem No>
    &lt;XRefSrcFile `"
					    (gen-file-name)
					    "'>
   > # end of XRef
"))))

;; (make-string parafmt stringfmt) creates a MIF string of FTag
;; stringfmt within a paragraph tagged as type parafmt, and handles
;; its content.  Pass parafmt if known; if not, pass the null string
;; "".
(define (make-string parafmt stringfmt)
  (sosofo-append (make formatting-instruction
		       data: (string-append "'>
   &lt;Font 
    &lt;FTag `"
					    stringfmt
					    "'>
    &lt;FLocked No>
   > # end of Font
   &lt;String `"))
		 (process-text parafmt stringfmt)))

;; (make-sub-para fmtname parentfmt) closes the current MIF paragraph,
;; creates a new one of type fmtname, and then re-opens one of type
;; parentfmt.
(define (make-sub-para fmtname
		       parentfmt)
  (sosofo-append (make formatting-instruction
		       data: "'>
  >
 > # end of Para
")
		 (make-para fmtname
			    #f)
		 (make formatting-instruction
		       data: (string-append (parafmt parentfmt)
					    "  &lt;ParaLine 
   &lt;String `"))))

;; (make-table fmtname cols) creates a MIF table with a TblTag of
;; fmtname and with cols columns.  cols is a string.
(define (make-table fmtname cols)
  (sosofo-append (make formatting-instruction
		       data: (string-append " &lt;Tbl 
  &lt;TblID "
					    (number->string (all-element-number))
					    ">
  &lt;TblFormat 
   &lt;TblTag `"
					    fmtname
					    "'>
"))
		 (make-columns)
		 (make formatting-instruction
		       data: (string-append "  > # end of TblFormat
  &lt;TblNumColumns "
					    cols
					    ">
"))
		 (make-table-content)
		 (make formatting-instruction
		       data: " > # end of Tbl
")))

;; (make-table-body) creates a MIF TblBody
(define (make-table-body)
  (sosofo-append (make formatting-instruction
		       data: "  &lt;TblBody 
")
		 (process-children)
		 (make formatting-instruction
		       data: "  > # end of TblBody
")))

;; (make-table-cell fmtname) creates a MIF Cell.  If no child of
;; a type listed in blocks is present, it also creates a paragraph of
;; type fmtname.
(define (make-table-cell fmtname)
  (sosofo-append (make formatting-instruction
		       data: "    &lt;Cell 
     &lt;CellContent 
      &lt;Notes 
")
		 (make-footnotes)
		 (make formatting-instruction
		       data: "      > # end of Notes
")
		 (if (node-list-empty? (get-children-by-type blocks
							     (current-node)))
		     (with-mode #f
				(make-para fmtname
					   #f))
		     (process-children))
		 (make formatting-instruction
		       data: "     > # end of CellContent
    > # end of Cell
")))

;; (make-table-foot) creates a MIF TblF.
(define (make-table-foot)
  (sosofo-append (make formatting-instruction
		       data: "  &lt;TblF 
")
		 (process-children)
		 (make formatting-instruction
		       data: "  > # end of TblF
")))

;; (make-table-head) creates a MIF TblH.
(define (make-table-head)
  (sosofo-append (make formatting-instruction
		       data: "  &lt;TblH 
")
		 (process-children)
		 (make formatting-instruction
		       data: "  > # end of TblH
")))

;; (make-table-ref) creates a MIF ATbl referring to a table.
(define (make-table-ref)
  (make formatting-instruction
	data: (string-append "   &lt;ATbl "
			     (number->string (all-element-number))
			     ">
")))

;; (make-table-row) creates a MIF Row.
(define (make-table-row)
  (sosofo-append (make formatting-instruction
		       data: "   &lt;Row 
")
		 (process-children)
		 (make formatting-instruction
		       data: "   > # end of Row
")))

;; (mif-char charname) ends the open String, inserts the MIF Character
;; charname, and resumes the String.
(define (mif-char charname)
  (make formatting-instruction
	data: (string-append "'>
   &lt;Char "
			     charname
			     ">
   &lt;String `")))

;; mif-nobreaks is a list of formats needed by these procedures as
;; part of the nobreaks list.
(define mif-nobreaks
  '("text-only"))

;; (new-line fmtname) ends and opens a String and a ParaLine within a
;; paragraph; for fmtnames listed in preserves, it creates a whole new
;; Para.
(define (new-line fmtname)
  (sosofo-append (make formatting-instruction
		       data: " '>
  >
")
		 (if (member (gi (current-node))
			     preserves)
		     (make formatting-instruction
			   data: (string-append " > # end of Para
 &lt;Para
  &lt;PgfTag `"
						fmtname
						"'>
"))
		     (empty-sosofo))
		 (make formatting-instruction
		       data: "  &lt;ParaLine 
   &lt;String `")))

;; (parafmt fmtname) creates a string with the start of a Para with a
;; PgfTag of fmtname.
(define (parafmt fmtname)
  (string-append " &lt;Para 
  &lt;PgfTag `"
		 fmtname
		 "'>
"))

;; (process-text fmtname stringfmt) handles the non-element content of
;; paragraphs.  Handles MIF reserved characters and subelements, and
;; calls handlers for PIs and SDATA entities.
(define (process-text fmtname stringfmt)
  (let p-t-loop ((this-node (node-list-first (children (current-node))))
		 (other-nodes (node-list-rest (children (current-node)))))
       (if (node-list-empty? this-node)
	   (empty-sosofo)
	   (sosofo-append (case (node-property 'class-name
					       this-node)
				;; handle special characters
				((data-char) (case (node-property 'char
								  this-node)
						   ;; carriage return
						   ((#\
) (new-line fmtname))
						   ;; single quote
						   ((#\') (make formatting-instruction
								data: "\\q"))
						   ;; greater-than sign
						   ((#\>) (make formatting-instruction
								data: "\\>"))
						   ;; back-slash
						   ((#\\) (make formatting-instruction
								data: "\\\\"))
						   (else (process-node-list this-node))))
				;; watch out for inline elements
				((element) (sosofo-append (process-node-list this-node)
							  (if (and (member (gi this-node)
									   inlines)
								   (not (equal? stringfmt
										"text-only")))
							      (make formatting-instruction
								    data: (string-append "'>
   &lt;Font 
    &lt;FTag `"
											 stringfmt
											 "'>
    &lt;FLocked No>
   > # end of Font
   &lt;String `"))
							      (empty-sosofo))))
				;; handle processing instructions
				((pi) (pi-handler this-node
						  parafmt
						  stringfmt))
				;; handle sdata entities
				((sdata) (sdata-handler this-node
							parafmt
							stringfmt)))
			  (if (node-list-empty? other-nodes)
			      (empty-sosofo)
			      (p-t-loop (node-list-first other-nodes)
					(node-list-rest other-nodes)))))))

;; (var-ref varname) makes a MIF reference to MIF variable varname.
(define (var-ref varname)
  (make formatting-instruction
	data: (string-append "'>
   &lt;Variable 
    &lt;VariableName `"
			     varname
			     "'>
   > # end of Variable
   &lt;String `")))

;; A mode for generating only text from elements.
(mode text-only
      (default (process-text "" "text-only")))

;; A default rule for tables mode.
(mode tables
      (default (empty-sosofo)))

<external-specification id="useful" document="useful">
